'use strict';

// YOU KNOW WHAT TO DO //

/**
 * each: Designed to loop over a collection, Array or Object, and applies the action 
 * Function to each value in the collection.
 * 
 * @param {Array or Object} collection The collection over which to iterate.
 * @param {Function} action The Function to be applied to each value in the 
 * collection
 */
function each(collection, action) {
    if(Array.isArray(collection)) {
        for(var i = 0; i < collection.length; i++) {
            action(collection[i], i, collection);
        }
    } else {
        for (var key in collection) {
            action(collection[key], key, collection);
        }
    }
}
module.exports.each = each;



/**
 * identity: Designed to return a value unchanged.
 * 
 * @param {Value} value to return unchanged
 * @return {Value} The value unchanged
*/
function identity(value){ //pass thru dummy function
   return value; 
}
module.exports.identity = identity; 


/**
 * typeof: Returns the data type of the value as a string.
 * 
 * @param {Value} value The value used to determine its data type.
 * @return {String} value The string representing the type of data. 
 * 
 * For ex: typeOf([]) // -> 'array'
 * ex: typeOf(null) // -> 'null'
 */
 function typeOf(value){
    if(Array.isArray(value)) return 'array';
    if(value === null) return 'null';
    if(value instanceof Date) return 'date';
    return typeof value; 
}
module.exports.typeof = typeOf;


/**
 * first: Takes an array and a number and will return the first 'n' number of 
 * elements in the array. If there is no number, it will return the first number
 * in the array. 
 * 
 * @param {Array} array The array used to return its elements.
 * @param {Number} n Amount that determines the number of elements to return from
 * the passed in array.
 * @return {[]} [] A new array containing the n amount of elements.
 */
function first(array, n) {
        if (!Array.isArray(array) || n < 0) return [];
        if (n === undefined || typeof n !== 'number' || n === 1) return array[0];
        return array.slice(0, n);
    }
    
    module.exports.first = first;
    
    
  /**
 * last: Takes an array and a number and will return the last 'n' number of 
 * elements in the array. If there is no number, it will return the last number
 * in the array. 
 * 
 * @param {Array} array The array used to return its elements.
 * @param {Number} n Amount that determines the number of elements to return 
 * from the passed in array.
 * @return {[]} [] A new array containing the n number of elements.
 */   
function last(array, n) {
    if (!Array.isArray(array) || n < 0) return [];
    if (n === undefined || typeof n !== 'number' || n === 1) return array[array.length - 1];
    if (n > array.length) return array;
    return array.slice(-n);

    }
module.exports.last = last;


/**
 * indexOf: Returns the index of the value in the array, if found. 
 * Returns -1 if the value is not in the array.
 * 
 * @param {Array} array The array to be iterated over in search of the value.
 * @param {Value} value The value of which to return its index within the array.
 * @return {Number} i The index of the value in the array.
 * @return {Number} -1 A falsy value indicating that the value is NOT in the
 * array.
 */
 
 function indexOf(array, value){  
    for(var i = 0; i < array.length; i++){
        if(array[i] === value){
            return i;
        }
    }    
    return -1;
}
module.exports.indexOf = indexOf;


/**
 * filter: Takes an array and a function and applies the function to the array. 
 * The function is used to test the elements of the array and return 
 * an array of the elements that have given a true result. 
 * 
 * @param {array} collection Takes an Array that holds the list 
 * of elements to be tested against
 * @param {Function} test The given function that takes a element, position, and 
 * array that is designed to return a boolean.
 * @return {[]} passed A new array containing the elements that have 'passed' 
 * the test.
 */
 
function filter(array, test) {
        var output = [];
        each(array, function (element, i, array) {
            if (test(element, i, array)) {
                output.push(element);
            }
        });
        return output;
    }
    
module.exports.filter = filter;


/**
 * filter: Takes an array and a function and applies the function to the array. 
 * The function is used to test the elements of the array and return 
 * an array of the elements that have given a false result. 
 * 
 * @param {Collection} collection Array or Object that holds the list of 
 * elements to be tested against
 * @param {Function} test The given function that takes a value, position, and 
 * collection that is designed to return a boolean.
 * @return {[]} passed A new array containing the elements that have 'failed' 
 * the test.
 */
function reject(array, test){
    return filter(array, function(value, position, collection){
        return !test(value, position, collection);
    });
}
module.exports.reject = reject;


/**
 * partition: Calls a function on each element in the array and 
 * returns an array of two sub-arrays. One sub-array containing true elements,
 * and the other sub-array containing false elements.
 * 
 * @param {Collection} collection The Array or Object used to apply the function
 * to all of its elements
 * @param {Function} test The test contains the passed in function used to 
 * determine which elements in the array are true and which are false.
 */
 
function partition(collection, test){
    return [filter(collection, test), reject(collection,test)];
}
module.exports.partition = partition;


/**
 * unique: Returns a new array of elements from the passed in array of elements 
 * with the duplicates removed.
 * 
 * @param {Array} array The array used to pull elements from for a new array.
 * @return {[]} newArray A new array with the duplicates removed.
 */
 function unique(array){
    var output = [];
    for(var i = 0; i < array.length; i++){
        if(indexOf(output, array[i]) === -1) output.push(array[i]); 
    }
    return output;
}
module.exports.unique = unique;


/**
 * map: Calls a function that transforms each element in a given collection and
 * returns them in a new array.
 * we pass in i and array becuase you might want to do look at the position in the array
  i and array are availible to you     
 * 
 * 
 * @param {Collection} Array or Object to pull elements from to apply the
 * transformation to.
 * @param {Function} transform The function that provides the action of 
 * transforming elements in the given collection
 * @return {[]} newArray an array containing the transformed elements
 */
 
function map(collection, transform){
    var output = [];
    each(collection, function(value, i, array){
        output.push(transform(value, i, array));
    });
    return output;
}
module.exports.map = map;


/**
 * pluck: Takes an array of objects and a property and returns every value at 
 * each property on every object in a new array
 * 
 * @param {Array} array The Array of objects.
 * @param {String} property The property/key that is on the objects in the array
 * and used to find the values.
 * @return {[]} output The array containing the values at every property
 */
function pluck(array, property){
    var output = [];
    map(array, function(element, i, array){
        output.push(element[property]);
    });
    return output;
}
module.exports.pluck = pluck;

/**
 * contains: Takes an array and a value that returns true if the value is in the
 * array, false otherwise
 * 
 *@param {Array} The Array used for its elements
 *@param {String} value The value passed in to be found in the given array
 * @return is returning false if the value isn't in the array
 */
function contains(array, value) {
        for (var i = 0; i < array.length; i++) {
            if (array[i] === value) {
                return true;
            }
        }
        return false;
    }
    
module.exports.contains = contains;    


/**
 * every: Takes a collection containing elements which will be tested against
 * to return true if all the elements pass the test or false if just
 * one does not pass. If a function is not passed in, one falsy 
 * element will return false and all truthy elements will return true.
 * 
 * @{Collection} The Array or Object to be iterated through
 * @{Test} the function thats going to test if the collection is true or false
 * if there is no input for test or function is not given, set it equal to undefined 
 *@param{collection[i], position, collection}
 * @param{collection[key], position, collection}
 * @return returns true if the test does not equal fasle 
 */

function every(collection, test) {
        if (test === undefined) test = identity();
        if (Array.isArray(collection)) {
            for (var i = 0; i < collection.length; i++) {
                if (!test(collection[i], i, collection)) return false;
            }
        } else {
            for (var key in collection) {
                if (!test(collection[key], key, collection)) return false;
            }
        }
        return true;
    }
module.exports.every = every;

/**
 * some: every: Takes a collection containing elements which will be tested 
 * against to return true if just one of the elements pass the function 
 * or false if none of them do. If a function is not passed in, one truthy 
 * element will return true and all falsy elements will return false.
 * 
 * @param {Collection} Array or object used for its elements
 * @param {Function} test function that takes the element, i, collection and 
 * returns a boolean.
 * @return returns false if the first test in the collection returns false
 */
 
function some(collection, test) {
        if (test === undefined) test = identity();
        if (Array.isArray(collection)) {
            for (var i = 0; i < collection.length; i++) {
                if (test(collection[i], i, collection)) return true;
            }
        } else {
            for (var key in collection) {
                if (test(collection[key], key, collection)) return true;
            }
        }
        return false;
    }
module.exports.some = some;

/**
 * reduce: This function will sum up all of the elements in the given 
 * collection. (as it iterates, it skips the first element. The first element is
 * used as the first 'operand' to be added to the first element that is iterated 
 * over.)
 * 
 * 
 * 
 * Start accumulating at the startValue, unless it's undefined, in that 
 * case, start accuummulating from the first value in the array, AND, 
 * start looping from the second value in the array.
 * we start looping from the second value in the array[0]
 * {for - loop} does the accumulation at the accumulate(accum, array[i], i, array);
 * @return returns the start value 
*/
 
    function reduce(array, accumulate, startValue) {
        
     var accum = startValue;
     var i = 0;
        if (accum === undefined) {
            accum = array[0];
            i = 1;
        }
        for(; i < array.length; i++) {
            accum = accumulate(accum, array[i], i, array);
        }
        return accum;
    }
module.exports.reduce = reduce;
    

/**
 * extend: Takes an unknown number of objects and it copies properties from each
 * object to the first one and then returns the updated object
 * 
 * @param {object1} Objects to copy properties to
 * @param {object2}
 * var copyFromObjects = Array.from(arguments).slice(1);
 * loop over all copy from objects 
 * pluck out copyFrom objects one at a time 
 * loop over copyFrom 
 * copy all values FROM copyFrom TO copyTo 
 * @return copyTO
 */
 
function extend(copyTo){
    var copyFromObjects = Array.prototype.slice.call(arguments, 1);

    for (var i = 0; i < copyFromObjects.length; i++) {
        var copyFrom = copyFromObjects[i];
            for (var key in copyFrom) {
             copyTo[key] = copyFrom[key];
               }
            }
          return copyTo;
            }
module.exports.extend = extend;        
        